#include <iostream>
#include <stdlib.h>
#include <vector>
#include <stdint.h>
#include <string.h>
#include <python2.7/Python.h>
#include "lsmp_con.hh"
#include "lsmp_prod.hh"
#include "checksum_crc32.hh"
#include "tconv.h"

#include "lsmp_rw.hh"

using namespace std;

ShmWriter::ShmWriter()
{
    partition_name.assign("");
}

const bool ShmWriter::connect(const char *name, const int num_buffers, const int blocksize, const int buffer_mode)
{
    partition_name.assign(name);
    partition = new LSMP_PROD(name, num_buffers, blocksize);
    if(!partition)
	return(false);
    if(!partition->valid()) {
	delete partition;
	return(false);
    }
    partition->keep(false);
    if(partition->lock(true)) {
	cout << "Warning: unable to lock partition: [" << partition->Error() << "]" << endl;
    }
    partition->bufmode(buffer_mode);
    return(true);
}

ShmWriter::~ShmWriter()
{
    delete partition;
}

const int ShmWriter::write_data(std::string write_data_bytes, const unsigned long timestamp) const
{
    const char *data;
    int length, max_length;
    const int mask = -1;

    const int flags = 0;
    data = (const char*) partition->get_buffer(flags);

    if(data == 0)
	return 1;

    length = write_data_bytes.length();
    max_length = partition->getBufferLength();
    if(length > max_length) {
	partition->return_buffer();
	return 1;
    }

    /* copy in data */
    memcpy((void*) data, (const void*) write_data_bytes.c_str(), length);

    partition->SetID(timestamp);

    write_data_bytes.assign((const char*) data, (const char*) data + length);
    
    partition->release(length, mask, flags);

    return 0;
}

/*
 * Taken from framexmit.cc and Dpush.C
 */
const bool ShmWriter::has_crc(const std::string data) const
{
    // Was a CRC even created for this? Check the byte at data[39]
    if(data[39]!=1)
	return(false);
    return(true);
}

const bool ShmWriter::check_crc(const std::string data) const
{
    if(!has_crc(data))
	return(false);

    const unsigned int length = data.length();
    const char* data_ptr = data.c_str();

    checksum_crc32 chk;
    uint32_t crc32;
    chk.add(data_ptr, length - 4);
    memcpy(&crc32, data_ptr+length-4, sizeof(crc32));
    if (crc32 != chk.result())
	return false;

    return(true);
}


ShmReader::ShmReader()
{
    partition_name.assign("");
}

const bool ShmReader::connect(const char *name)
{
    partition_name.assign(name);
    partition = new LSMP_CON(name);
    if(!partition)
	return(false);
    if(!partition->isConnected()) {
	delete partition;
	return(false);
    }
    partition->setTimeout(-1);
    return(true);
}

ShmReader::~ShmReader()
{
    delete partition;
}

void ShmReader::read_data(int& read_data_ret, std::string& read_data_bytes,
			  unsigned long& read_data_timestamp)
{
    char *data;
    int length;

    const int flags = 0;

    //
    // Since the next statement will likely block waiting for the memory
    // to be filled, we need to tell the Python interpreter than it can
    // go do things elsewhere. For more information see e.g.
    // https://opensource.com/article/17/4/grok-gil
    // http://www.dabeaz.com/GIL/
    // various youtube videos from David Beazley, e.g.
    // https://www.youtube.com/watch?v=Obt-vMVdM8s
    Py_BEGIN_ALLOW_THREADS
    data = (char*) partition->get_buffer(flags);
    Py_END_ALLOW_THREADS

    if(data == 0) {
	read_data_ret = 1;
	return;
    }

    length = partition->getLength();

    read_data_timestamp = partition->getEvtID();

    read_data_bytes.assign((const char*) data, (const char*) data + length);
    
    partition->free_buffer();

    read_data_ret=0;
}

/*
 * Taken from framexmit.cc and Dpush.C
 */
const bool ShmReader::has_crc(const std::string data)
{
    // Was a CRC even created for this? Check the byte at data[39]
    if(data[39]!=1)
	return(false);
    return(true);
}

const bool ShmReader::check_crc(const std::string data)
{
    if(!has_crc(data))
	return(false);

    unsigned int length = data.length();
    const char* data_ptr = data.c_str();

    checksum_crc32 chk;
    uint32_t crc32;
    chk.add(data_ptr, length - 4);
    memcpy(&crc32, data_ptr+length-4, sizeof(crc32));
    if (crc32 != chk.result())
	return false;

    return(true);
}


class LSmp_Kill : private LSMP {
public:
    LSmp_Kill(const char* name);
    ~LSmp_Kill(void);
};

LSmp_Kill::LSmp_Kill(const char* name) : LSMP()
{
    if (name) Zeuss(name);
}

LSmp_Kill::~LSmp_Kill(void) {}

int ShmKill::kill(const char *name)
{
    // create an LSmp_Kill object wrapping, which will then get killed when
    // its destructor is called
    LSmp_Kill goner(name);
    return(0);
}

